@@ -18,6 +18,7 @@ from urllib.parse import unquote
 from urllib.request import parse_http_list as _parse_list_header
 
 from ._internal import _dt_as_utc
+from ._internal import _plain_int
 
 if t.TYPE_CHECKING:
     from _typeshed.wsgi import WSGIEnvironment
@@ -153,16 +154,16 @@ def quote_header_value(
         The value is quoted if it is the empty string.
 
     .. versionchanged:: 2.3
-        Passing bytes is deprecated and will not be supported in Werkzeug 2.4.
+        Passing bytes is deprecated and will not be supported in Werkzeug 3.0.
 
     .. versionchanged:: 2.3
-        The ``extra_chars`` parameter is deprecated and will be removed in Werkzeug 2.4.
+        The ``extra_chars`` parameter is deprecated and will be removed in Werkzeug 3.0.
 
     .. versionadded:: 0.5
     """
     if isinstance(value, bytes):
         warnings.warn(
-            "Passing bytes is deprecated and will not be supported in Werkzeug 2.4.",
+            "Passing bytes is deprecated and will not be supported in Werkzeug 3.0.",
             DeprecationWarning,
             stacklevel=2,
         )
@@ -171,7 +172,7 @@ def quote_header_value(
     if extra_chars is not None:
         warnings.warn(
             "The 'extra_chars' parameter is deprecated and will be"
-            " removed in Werkzeug 2.4.",
+            " removed in Werkzeug 3.0.",
             DeprecationWarning,
             stacklevel=2,
         )
@@ -203,12 +204,12 @@ def unquote_header_value(value: str, is_filename: bool | None = None) -> str:
     :param value: The header value to unquote.
 
     .. versionchanged:: 2.3
-        The ``is_filename`` parameter is deprecated and will be removed in Werkzeug 2.4.
+        The ``is_filename`` parameter is deprecated and will be removed in Werkzeug 3.0.
     """
     if is_filename is not None:
         warnings.warn(
             "The 'is_filename' parameter is deprecated and will be"
-            " removed in Werkzeug 2.4.",
+            " removed in Werkzeug 3.0.",
             DeprecationWarning,
             stacklevel=2,
         )
@@ -298,7 +299,7 @@ def dump_header(
     :param iterable: The items to create a header from.
 
     .. versionchanged:: 2.3
-        The ``allow_token`` parameter is deprecated and will be removed in Werkzeug 2.4.
+        The ``allow_token`` parameter is deprecated and will be removed in Werkzeug 3.0.
 
     .. versionchanged:: 2.2.3
         If a key ends with ``*``, its value will not be quoted.
@@ -306,7 +307,7 @@ def dump_header(
     if allow_token is not None:
         warnings.warn(
             "'The 'allow_token' parameter is deprecated and will be"
-            " removed in Werkzeug 2.4.",
+            " removed in Werkzeug 3.0.",
             DeprecationWarning,
             stacklevel=2,
         )
@@ -394,10 +395,10 @@ def parse_dict_header(value: str, cls: type[dict] | None = None) -> dict[str, st
         Added support for ``key*=charset''value`` encoded items.
 
     .. versionchanged:: 2.3
-        Passing bytes is deprecated, support will be removed in Werkzeug 2.4.
+        Passing bytes is deprecated, support will be removed in Werkzeug 3.0.
 
     .. versionchanged:: 2.3
-        The ``cls`` argument is deprecated and will be removed in Werkzeug 2.4.
+        The ``cls`` argument is deprecated and will be removed in Werkzeug 3.0.
 
     .. versionchanged:: 0.9
        The ``cls`` argument was added.
@@ -406,7 +407,7 @@ def parse_dict_header(value: str, cls: type[dict] | None = None) -> dict[str, st
         cls = dict
     else:
         warnings.warn(
-            "The 'cls' parameter is deprecated and will be removed in Werkzeug 2.4.",
+            "The 'cls' parameter is deprecated and will be removed in Werkzeug 3.0.",
             DeprecationWarning,
             stacklevel=2,
         )
@@ -415,7 +416,7 @@ def parse_dict_header(value: str, cls: type[dict] | None = None) -> dict[str, st
 
     if isinstance(value, bytes):
         warnings.warn(
-            "Passing bytes is deprecated and will be removed in Werkzeug 2.4.",
+            "Passing bytes is deprecated and will be removed in Werkzeug 3.0.",
             DeprecationWarning,
             stacklevel=2,
         )
@@ -612,6 +613,7 @@ def parse_options_header(value: str | None) -> tuple[str, dict[str, str]]:
     return value, options
 
 
+_q_value_re = re.compile(r"-?\d+(\.\d+)?", re.ASCII)
 _TAnyAccept = t.TypeVar("_TAnyAccept", bound="ds.Accept")
 
 
@@ -654,13 +656,15 @@ def parse_accept_header(
         item, options = parse_options_header(item)
 
         if "q" in options:
-            try:
-                # pop q, remaining options are reconstructed
-                q = float(options.pop("q"))
-            except ValueError:
+            # pop q, remaining options are reconstructed
+            q_str = options.pop("q").strip()
+
+            if _q_value_re.fullmatch(q_str) is None:
                 # ignore an invalid q
                 continue
 
+            q = float(q_str)
+
             if q < 0 or q > 1:
                 # ignore an invalid q
                 continue
@@ -823,7 +827,7 @@ def parse_authorization_header(
     :return: a :class:`~werkzeug.datastructures.Authorization` object or `None`.
 
     .. deprecated:: 2.3
-        Will be removed in Werkzeug 2.4. Use :meth:`.Authorization.from_header` instead.
+        Will be removed in Werkzeug 3.0. Use :meth:`.Authorization.from_header` instead.
     """
     from .datastructures import Authorization
 
@@ -850,7 +854,7 @@ def parse_www_authenticate_header(
     :return: a :class:`~werkzeug.datastructures.WWWAuthenticate` object.
 
     .. deprecated:: 2.3
-        Will be removed in Werkzeug 2.4. Use :meth:`.WWWAuthenticate.from_header`
+        Will be removed in Werkzeug 3.0. Use :meth:`.WWWAuthenticate.from_header`
         instead.
     """
     from .datastructures.auth import WWWAuthenticate
@@ -914,7 +918,7 @@ def parse_range_header(
             if last_end < 0:
                 return None
             try:
-                begin = int(item)
+                begin = _plain_int(item)
             except ValueError:
                 return None
             end = None
@@ -925,7 +929,7 @@ def parse_range_header(
             end_str = end_str.strip()
 
             try:
-                begin = int(begin_str)
+                begin = _plain_int(begin_str)
             except ValueError:
                 return None
 
@@ -933,7 +937,7 @@ def parse_range_header(
                 return None
             if end_str:
                 try:
-                    end = int(end_str) + 1
+                    end = _plain_int(end_str) + 1
                 except ValueError:
                     return None
 
@@ -976,7 +980,7 @@ def parse_content_range_header(
         length = None
     else:
         try:
-            length = int(length_str)
+            length = _plain_int(length_str)
         except ValueError:
             return None
 
@@ -990,8 +994,8 @@ def parse_content_range_header(
 
     start_str, stop_str = rng.split("-", 1)
     try:
-        start = int(start_str)
-        stop = int(stop_str) + 1
+        start = _plain_int(start_str)
+        stop = _plain_int(stop_str) + 1
     except ValueError:
         return None
 
@@ -1298,7 +1302,7 @@ def parse_cookie(
 
     .. versionchanged:: 2.3
         Passing bytes, and the ``charset`` and ``errors`` parameters, are deprecated and
-        will be removed in Werkzeug 2.4.
+        will be removed in Werkzeug 3.0.
 
     .. versionchanged:: 1.0
         Returns a :class:`MultiDict` instead of a ``TypeConversionDict``.
@@ -1311,7 +1315,7 @@ def parse_cookie(
         cookie = header.get("HTTP_COOKIE")
     elif isinstance(header, bytes):
         warnings.warn(
-            "Passing bytes is deprecated and will not be supported in Werkzeug 2.4.",
+            "Passing bytes is deprecated and will not be supported in Werkzeug 3.0.",
             DeprecationWarning,
             stacklevel=2,
         )
@@ -1369,7 +1373,7 @@ def dump_cookie(
     :param path: limits the cookie to a given path, per default it will
                  span the whole domain.
     :param domain: Use this if you want to set a cross-domain cookie. For
-                   example, ``domain=".example.com"`` will set a cookie
+                   example, ``domain="example.com"`` will set a cookie
                    that is readable by the domain ``www.example.com``,
                    ``foo.example.com`` etc. Otherwise, a cookie will only
                    be readable by the domain that set it.
@@ -1388,7 +1392,7 @@ def dump_cookie(
 
     .. _`cookie`: http://browsercookielimits.squawky.net/
 
-    .. versionchanged:: 2.3.2
+    .. versionchanged:: 2.3.3
         The ``path`` parameter is ``/`` by default.
 
     .. versionchanged:: 2.3.1
@@ -1403,7 +1407,7 @@ def dump_cookie(
 
     .. versionchanged:: 2.3
         Passing bytes, and the ``charset`` parameter, are deprecated and will be removed
-        in Werkzeug 2.4.
+        in Werkzeug 3.0.
 
     .. versionchanged:: 1.0.0
         The string ``'None'`` is accepted for ``samesite``.
@@ -1411,7 +1415,7 @@ def dump_cookie(
     if charset is not None:
         warnings.warn(
             "The 'charset' parameter is deprecated and will be removed"
-            " in Werkzeug 2.4.",
+            " in Werkzeug 3.0.",
             DeprecationWarning,
             stacklevel=2,
         )
@@ -1421,7 +1425,7 @@ def dump_cookie(
     if isinstance(key, bytes):
         warnings.warn(
             "The 'key' parameter must be a string. Bytes are deprecated"
-            " and will not be supported in Werkzeug 2.4.",
+            " and will not be supported in Werkzeug 3.0.",
             DeprecationWarning,
             stacklevel=2,
         )
@@ -1430,7 +1434,7 @@ def dump_cookie(
     if isinstance(value, bytes):
         warnings.warn(
             "The 'value' parameter must be a string. Bytes are"
-            " deprecated and will not be supported in Werkzeug 2.4.",
+            " deprecated and will not be supported in Werkzeug 3.0.",
             DeprecationWarning,
             stacklevel=2,
         )
