@@ -1,13 +1,15 @@
+from __future__ import annotations
+
 import ast
 import re
 import typing as t
 from dataclasses import dataclass
 from string import Template
 from types import CodeType
+from urllib.parse import quote
 
-from .._internal import _to_bytes
-from ..urls import url_encode
-from ..urls import url_quote
+from ..datastructures import iter_multi_items
+from ..urls import _urlencode
 from .converters import ValidationError
 
 if t.TYPE_CHECKING:
@@ -17,9 +19,9 @@ if t.TYPE_CHECKING:
 
 class Weighting(t.NamedTuple):
     number_static_weights: int
-    static_weights: t.List[t.Tuple[int, int]]
+    static_weights: list[tuple[int, int]]
     number_argument_weights: int
-    argument_weights: t.List[int]
+    argument_weights: list[int]
 
 
 @dataclass
@@ -36,22 +38,28 @@ class RulePart:
     content: str
     final: bool
     static: bool
+    suffixed: bool
     weight: Weighting
 
 
 _part_re = re.compile(
     r"""
-    (?P<static>[^<]*)                             # static rule data
     (?:
-      <
+        (?P<slash>/)                                 # a slash
+      |
+        (?P<static>[^</]+)                           # static rule data
+      |
         (?:
-          (?P<converter>[a-zA-Z_][a-zA-Z0-9_]*)   # converter name
-          (?:\((?P<arguments>.*?)\))?             # converter arguments
-          \:                                      # variable delimiter
-        )?
-        (?P<variable>[a-zA-Z_][a-zA-Z0-9_]*)      # variable name
-       >
-    )?
+          <
+            (?:
+              (?P<converter>[a-zA-Z_][a-zA-Z0-9_]*)   # converter name
+              (?:\((?P<arguments>.*?)\))?             # converter arguments
+              :                                       # variable delimiter
+            )?
+            (?P<variable>[a-zA-Z_][a-zA-Z0-9_]*)      # variable name
+           >
+        )
+    )
     """,
     re.VERBOSE,
 )
@@ -87,7 +95,7 @@ def _find(value: str, target: str, pos: int) -> int:
         return len(value)
 
 
-def _pythonize(value: str) -> t.Union[None, bool, int, float, str]:
+def _pythonize(value: str) -> None | bool | int | float | str:
     if value in _PYTHON_CONSTANTS:
         return _PYTHON_CONSTANTS[value]
     for convert in int, float:
@@ -100,7 +108,7 @@ def _pythonize(value: str) -> t.Union[None, bool, int, float, str]:
     return str(value)
 
 
-def parse_converter_args(argstr: str) -> t.Tuple[t.Tuple, t.Dict[str, t.Any]]:
+def parse_converter_args(argstr: str) -> tuple[t.Tuple, dict[str, t.Any]]:
     argstr += ","
     args = []
     kwargs = {}
@@ -125,7 +133,7 @@ class RuleFactory:
     be added by subclassing `RuleFactory` and overriding `get_rules`.
     """
 
-    def get_rules(self, map: "Map") -> t.Iterable["Rule"]:
+    def get_rules(self, map: Map) -> t.Iterable[Rule]:
         """Subclasses of `RuleFactory` have to override this method and return
         an iterable of rules."""
         raise NotImplementedError()
@@ -154,7 +162,7 @@ class Subdomain(RuleFactory):
         self.subdomain = subdomain
         self.rules = rules
 
-    def get_rules(self, map: "Map") -> t.Iterator["Rule"]:
+    def get_rules(self, map: Map) -> t.Iterator[Rule]:
         for rulefactory in self.rules:
             for rule in rulefactory.get_rules(map):
                 rule = rule.empty()
@@ -180,7 +188,7 @@ class Submount(RuleFactory):
         self.path = path.rstrip("/")
         self.rules = rules
 
-    def get_rules(self, map: "Map") -> t.Iterator["Rule"]:
+    def get_rules(self, map: Map) -> t.Iterator[Rule]:
         for rulefactory in self.rules:
             for rule in rulefactory.get_rules(map):
                 rule = rule.empty()
@@ -205,7 +213,7 @@ class EndpointPrefix(RuleFactory):
         self.prefix = prefix
         self.rules = rules
 
-    def get_rules(self, map: "Map") -> t.Iterator["Rule"]:
+    def get_rules(self, map: Map) -> t.Iterator[Rule]:
         for rulefactory in self.rules:
             for rule in rulefactory.get_rules(map):
                 rule = rule.empty()
@@ -232,10 +240,10 @@ class RuleTemplate:
     replace the placeholders in all the string parameters.
     """
 
-    def __init__(self, rules: t.Iterable["Rule"]) -> None:
+    def __init__(self, rules: t.Iterable[Rule]) -> None:
         self.rules = list(rules)
 
-    def __call__(self, *args: t.Any, **kwargs: t.Any) -> "RuleTemplateFactory":
+    def __call__(self, *args: t.Any, **kwargs: t.Any) -> RuleTemplateFactory:
         return RuleTemplateFactory(self.rules, dict(*args, **kwargs))
 
 
@@ -247,12 +255,12 @@ class RuleTemplateFactory(RuleFactory):
     """
 
     def __init__(
-        self, rules: t.Iterable[RuleFactory], context: t.Dict[str, t.Any]
+        self, rules: t.Iterable[RuleFactory], context: dict[str, t.Any]
     ) -> None:
         self.rules = rules
         self.context = context
 
-    def get_rules(self, map: "Map") -> t.Iterator["Rule"]:
+    def get_rules(self, map: Map) -> t.Iterator[Rule]:
         for rulefactory in self.rules:
             for rule in rulefactory.get_rules(map):
                 new_defaults = subdomain = None
@@ -433,25 +441,26 @@ class Rule(RuleFactory):
     def __init__(
         self,
         string: str,
-        defaults: t.Optional[t.Mapping[str, t.Any]] = None,
-        subdomain: t.Optional[str] = None,
-        methods: t.Optional[t.Iterable[str]] = None,
+        defaults: t.Mapping[str, t.Any] | None = None,
+        subdomain: str | None = None,
+        methods: t.Iterable[str] | None = None,
         build_only: bool = False,
-        endpoint: t.Optional[str] = None,
-        strict_slashes: t.Optional[bool] = None,
-        merge_slashes: t.Optional[bool] = None,
-        redirect_to: t.Optional[t.Union[str, t.Callable[..., str]]] = None,
+        endpoint: str | None = None,
+        strict_slashes: bool | None = None,
+        merge_slashes: bool | None = None,
+        redirect_to: str | t.Callable[..., str] | None = None,
         alias: bool = False,
-        host: t.Optional[str] = None,
+        host: str | None = None,
         websocket: bool = False,
     ) -> None:
         if not string.startswith("/"):
-            raise ValueError("urls must start with a leading slash")
+            raise ValueError(f"URL rule '{string}' must start with a slash.")
+
         self.rule = string
         self.is_leaf = not string.endswith("/")
         self.is_branch = string.endswith("/")
 
-        self.map: "Map" = None  # type: ignore
+        self.map: Map = None  # type: ignore
         self.strict_slashes = strict_slashes
         self.merge_slashes = merge_slashes
         self.subdomain = subdomain
@@ -484,11 +493,11 @@ class Rule(RuleFactory):
         else:
             self.arguments = set()
 
-        self._converters: t.Dict[str, "BaseConverter"] = {}
-        self._trace: t.List[t.Tuple[bool, str]] = []
-        self._parts: t.List[RulePart] = []
+        self._converters: dict[str, BaseConverter] = {}
+        self._trace: list[tuple[bool, str]] = []
+        self._parts: list[RulePart] = []
 
-    def empty(self) -> "Rule":
+    def empty(self) -> Rule:
         """
         Return an unbound copy of this rule.
 
@@ -525,7 +534,7 @@ class Rule(RuleFactory):
             host=self.host,
         )
 
-    def get_rules(self, map: "Map") -> t.Iterator["Rule"]:
+    def get_rules(self, map: Map) -> t.Iterator[Rule]:
         yield self
 
     def refresh(self) -> None:
@@ -536,7 +545,7 @@ class Rule(RuleFactory):
         """
         self.bind(self.map, rebind=True)
 
-    def bind(self, map: "Map", rebind: bool = False) -> None:
+    def bind(self, map: Map, rebind: bool = False) -> None:
         """Bind the url to a map and create a regular expression based on
         the information from the rule itself and the defaults from the map.
 
@@ -559,7 +568,7 @@ class Rule(RuleFactory):
         converter_name: str,
         args: t.Tuple,
         kwargs: t.Mapping[str, t.Any],
-    ) -> "BaseConverter":
+    ) -> BaseConverter:
         """Looks up the converter for the given parameter.
 
         .. versionadded:: 0.9
@@ -569,40 +578,38 @@ class Rule(RuleFactory):
         return self.map.converters[converter_name](self.map, *args, **kwargs)
 
     def _encode_query_vars(self, query_vars: t.Mapping[str, t.Any]) -> str:
-        return url_encode(
-            query_vars,
-            charset=self.map.charset,
-            sort=self.map.sort_parameters,
-            key=self.map.sort_key,
-        )
+        items: t.Iterable[tuple[str, str]] = iter_multi_items(query_vars)
+
+        if self.map.sort_parameters:
+            items = sorted(items, key=self.map.sort_key)
+
+        return _urlencode(items, encoding=self.map.charset)
 
     def _parse_rule(self, rule: str) -> t.Iterable[RulePart]:
-        pos = 0
-        endpos = _find(rule, "/", pos)
         content = ""
         static = True
         argument_weights = []
-        static_weights = []
+        static_weights: list[tuple[int, int]] = []
         final = False
-        static_parts = 0
+        convertor_number = 0
 
-        while pos <= len(rule):
-            match = _part_re.match(rule, pos, endpos)
-            assert match is not None
+        pos = 0
+        while pos < len(rule):
+            match = _part_re.match(rule, pos)
+            if match is None:
+                raise ValueError(f"malformed url rule: {rule!r}")
 
             data = match.groupdict()
-            if data["static"]:
-                # Only static parts beyond the `/` count towards the
-                # weighting (more static parts mean the rule should be
-                # weighted first).
-                if data["static"] != "/":
-                    static_weights.append((static_parts, -len(data["static"])))
-                    static_parts += 1
+            if data["static"] is not None:
+                static_weights.append((len(static_weights), -len(data["static"])))
                 self._trace.append((False, data["static"]))
+                content += data["static"] if static else re.escape(data["static"])
 
             if data["variable"] is not None:
+                if static:
+                    # Switching content to represent regex, hence the need to escape
+                    content = re.escape(content)
                 static = False
-                content += re.escape(data["static"])
                 c_args, c_kwargs = parse_converter_args(data["arguments"] or "")
                 convobj = self.get_converter(
                     data["variable"], data["converter"] or "default", c_args, c_kwargs
@@ -610,35 +617,66 @@ class Rule(RuleFactory):
                 self._converters[data["variable"]] = convobj
                 self.arguments.add(data["variable"])
                 if not convobj.part_isolating:
-                    endpos = len(rule)
                     final = True
-                content += f"({convobj.regex})"
+                content += f"(?P<__werkzeug_{convertor_number}>{convobj.regex})"
+                convertor_number += 1
                 argument_weights.append(convobj.weight)
                 self._trace.append((True, data["variable"]))
-            else:
-                content += data["static"] if static else re.escape(data["static"])
+
+            if data["slash"] is not None:
+                self._trace.append((False, "/"))
+                if final:
+                    content += "/"
+                else:
+                    if not static:
+                        content += r"\Z"
+                    weight = Weighting(
+                        -len(static_weights),
+                        static_weights,
+                        -len(argument_weights),
+                        argument_weights,
+                    )
+                    yield RulePart(
+                        content=content,
+                        final=final,
+                        static=static,
+                        suffixed=False,
+                        weight=weight,
+                    )
+                    content = ""
+                    static = True
+                    argument_weights = []
+                    static_weights = []
+                    final = False
+                    convertor_number = 0
 
             pos = match.end()
-            if pos == endpos:
-                if pos < len(rule) and rule[pos] == "/":
-                    self._trace.append((False, "/"))
-                pos += 1
-                weight = Weighting(
-                    -len(static_weights),
-                    static_weights,
-                    -len(argument_weights),
-                    argument_weights,
-                )
-                if not static:
-                    content += r"\Z"
-                yield RulePart(
-                    content=content, final=final, static=static, weight=weight
-                )
-                content = ""
-                static = True
-                argument_weights = []
-                static_weights = []
-                endpos = _find(rule, "/", pos)
+
+        suffixed = False
+        if final and content[-1] == "/":
+            # If a converter is part_isolating=False (matches slashes) and ends with a
+            # slash, augment the regex to support slash redirects.
+            suffixed = True
+            content = content[:-1] + "(?<!/)(/?)"
+        if not static:
+            content += r"\Z"
+        weight = Weighting(
+            -len(static_weights),
+            static_weights,
+            -len(argument_weights),
+            argument_weights,
+        )
+        yield RulePart(
+            content=content,
+            final=final,
+            static=static,
+            suffixed=suffixed,
+            weight=weight,
+        )
+        if suffixed:
+            yield RulePart(
+                content="", final=False, static=True, suffixed=False, weight=weight
+            )
 
     def compile(self) -> None:
         """Compiles the regular expression and stores it."""
@@ -651,31 +689,42 @@ class Rule(RuleFactory):
         self._parts = []
         self._trace = []
         self._converters = {}
-        self._parts.extend(self._parse_rule(domain_rule))
+        if domain_rule == "":
+            self._parts = [
+                RulePart(
+                    content="",
+                    final=False,
+                    static=True,
+                    suffixed=False,
+                    weight=Weighting(0, [], 0, []),
+                )
+            ]
+        else:
+            self._parts.extend(self._parse_rule(domain_rule))
         self._trace.append((False, "|"))
         rule = self.rule
         if self.merge_slashes:
             rule = re.sub("/{2,}?", "/", self.rule)
         self._parts.extend(self._parse_rule(rule))
 
-        self._build: t.Callable[..., t.Tuple[str, str]]
+        self._build: t.Callable[..., tuple[str, str]]
         self._build = self._compile_builder(False).__get__(self, None)
-        self._build_unknown: t.Callable[..., t.Tuple[str, str]]
+        self._build_unknown: t.Callable[..., tuple[str, str]]
         self._build_unknown = self._compile_builder(True).__get__(self, None)
 
     @staticmethod
-    def _get_func_code(code: CodeType, name: str) -> t.Callable[..., t.Tuple[str, str]]:
-        globs: t.Dict[str, t.Any] = {}
-        locs: t.Dict[str, t.Any] = {}
+    def _get_func_code(code: CodeType, name: str) -> t.Callable[..., tuple[str, str]]:
+        globs: dict[str, t.Any] = {}
+        locs: dict[str, t.Any] = {}
         exec(code, globs, locs)
         return locs[name]  # type: ignore
 
     def _compile_builder(
         self, append_unknown: bool = True
-    ) -> t.Callable[..., t.Tuple[str, str]]:
+    ) -> t.Callable[..., tuple[str, str]]:
         defaults = self.defaults or {}
-        dom_ops: t.List[t.Tuple[bool, str]] = []
-        url_ops: t.List[t.Tuple[bool, str]] = []
+        dom_ops: list[tuple[bool, str]] = []
+        url_ops: list[tuple[bool, str]] = []
 
         opl = dom_ops
         for is_dynamic, data in self._trace:
@@ -689,8 +738,12 @@ class Rule(RuleFactory):
                 data = self._converters[data].to_url(defaults[data])
                 opl.append((False, data))
             elif not is_dynamic:
+                # safe = https://url.spec.whatwg.org/#url-path-segment-string
                 opl.append(
-                    (False, url_quote(_to_bytes(data, self.map.charset), safe="/:|+"))
+                    (
+                        False,
+                        quote(data, safe="!$&'()*+,/:;=@", encoding=self.map.charset),
+                    )
                 )
             else:
                 opl.append((True, data))
@@ -700,17 +753,17 @@ class Rule(RuleFactory):
             ret.args = [ast.Name(str(elem), ast.Load())]  # type: ignore  # str for py2
             return ret
 
-        def _parts(ops: t.List[t.Tuple[bool, str]]) -> t.List[ast.AST]:
+        def _parts(ops: list[tuple[bool, str]]) -> list[ast.AST]:
             parts = [
-                _convert(elem) if is_dynamic else ast.Str(s=elem)
+                _convert(elem) if is_dynamic else ast.Constant(elem)
                 for is_dynamic, elem in ops
             ]
-            parts = parts or [ast.Str("")]
+            parts = parts or [ast.Constant("")]
             # constant fold
             ret = [parts[0]]
             for p in parts[1:]:
-                if isinstance(p, ast.Str) and isinstance(ret[-1], ast.Str):
-                    ret[-1] = ast.Str(ret[-1].s + p.s)
+                if isinstance(p, ast.Constant) and isinstance(ret[-1], ast.Constant):
+                    ret[-1] = ast.Constant(ret[-1].value + p.value)
                 else:
                     ret.append(p)
             return ret
@@ -723,7 +776,7 @@ class Rule(RuleFactory):
             body = [_IF_KWARGS_URL_ENCODE_AST]
             url_parts.extend(_URL_ENCODE_AST_NAMES)
 
-        def _join(parts: t.List[ast.AST]) -> ast.AST:
+        def _join(parts: list[ast.AST]) -> ast.AST:
             if len(parts) == 1:  # shortcut
                 return parts[0]
             return ast.JoinedStr(parts)
@@ -746,11 +799,11 @@ class Rule(RuleFactory):
             func_ast.args.args.append(ast.arg(arg, None))
         func_ast.args.kwarg = ast.arg(".kwargs", None)
         for _ in kargs:
-            func_ast.args.defaults.append(ast.Str(""))
+            func_ast.args.defaults.append(ast.Constant(""))
         func_ast.body = body
 
-        # use `ast.parse` instead of `ast.Module` for better portability
-        # Python 3.8 changes the signature of `ast.Module`
+        # Use `ast.parse` instead of `ast.Module` for better portability, since the
+        # signature of `ast.Module` can change.
         module = ast.parse("")
         module.body = [func_ast]
 
@@ -761,18 +814,18 @@ class Rule(RuleFactory):
             if "lineno" in node._attributes:
                 node.lineno = 1
             if "end_lineno" in node._attributes:
-                node.end_lineno = node.lineno  # type: ignore[attr-defined]
+                node.end_lineno = node.lineno
             if "col_offset" in node._attributes:
                 node.col_offset = 0
             if "end_col_offset" in node._attributes:
-                node.end_col_offset = node.col_offset  # type: ignore[attr-defined]
+                node.end_col_offset = node.col_offset
 
         code = compile(module, "<werkzeug routing>", "exec")
         return self._get_func_code(code, func_ast.name)
 
     def build(
         self, values: t.Mapping[str, t.Any], append_unknown: bool = True
-    ) -> t.Optional[t.Tuple[str, str]]:
+    ) -> tuple[str, str] | None:
         """Assembles the relative url for that rule and the subdomain.
         If building doesn't work for some reasons `None` is returned.
 
@@ -786,7 +839,7 @@ class Rule(RuleFactory):
         except ValidationError:
             return None
 
-    def provides_defaults_for(self, rule: "Rule") -> bool:
+    def provides_defaults_for(self, rule: Rule) -> bool:
         """Check if this rule has defaults for a given rule.
 
         :internal:
@@ -800,7 +853,7 @@ class Rule(RuleFactory):
         )
 
     def suitable_for(
-        self, values: t.Mapping[str, t.Any], method: t.Optional[str] = None
+        self, values: t.Mapping[str, t.Any], method: str | None = None
     ) -> bool:
         """Check if the dict of values has enough data for url generation.
 
@@ -832,7 +885,7 @@ class Rule(RuleFactory):
 
         return True
 
-    def build_compare_key(self) -> t.Tuple[int, int, int]:
+    def build_compare_key(self) -> tuple[int, int, int]:
         """The build compare key for sorting.
 
         :internal:
